//==============================================================================
//
// Projet  : FPROFIL
//           Classe de fichier de profil et fonctions d'acces
//
// Version : 1.00  Juillet 1999
// Auteur  : Denis ZIEGLER  denis.ziegler@worldonline.fr
//
// Fichier : FPROFIL.CPP
//
// Portabilite : C++ ANSI (DOS, Windows, Unix)
//
// (C) Copyright 1999 - ZIEGLER Denis
// Diffusion/modification autorisee selon les termes de la licence d'utilisation
//==============================================================================


#include <stdlib.h>

#include "FProfil.h"



//===== Classe de fichier de profil ============================================

//--- Constructeur -------------------------------------------------------------
FichProf::FichProf(char *nomfich, int _lectauto, int _ecrauto) : FichTxt(nomfich)
{
  profil = sectcour = NULL;
  entrcour = NULL;
  modif = 0;
  lectauto = _lectauto;
  ecrauto = _ecrauto;
  if (lectauto)
    LireProfil();  //lecture automatique du fichier de profil
}


//--- Destructeur -------------------------------------------------------------
FichProf::~FichProf(void)
{
  if (ecrauto && modif)
    EcrireProfil();     //ecriture automatique du fichier si profil modifie
  DesallouerProfil();   //desallocation des structures de donnees en memoire
}


//--- Lecture d'une entree dans le profil --------------------------------------
int FichProf::LireCh(char *sect,    //nom de la section a lire
                     char *entr,    //nom de l'entree a lire
                     char *defaut,  //valeur par defaut
                     char *dest,    //chaine destination
                     int dim)       //taille maxi de dest
{
  int trouve=0;
  int i=0;
  int imax;
  char chtemp[256];  //chaine temporaire

  // Recherche de la section dans la liste chainee
  sectcour = profil;
  while ((sectcour) && (!trouve))
    if (sectcour->comment)
      sectcour = sectcour->suivt;
    else
      if (!stricomp(sectcour->id, sect)) //comparaison avec la chaine recherchee
        trouve = 1;                         //(sans distinction min/maj)
      else
        sectcour = sectcour->suivt;
  // Section non trouvee --> copie de la valeur par defaut
  if (!trouve)
    AjouteCh(dest, defaut, i, dim);
  else
    // Section trouvee
    //   a) Si pas d'entree a chercher --> recopie de toute la section
    if (!entr) {
      entrcour = sectcour->entr;
      imax = dim - 1;
      while ((entrcour) && (i < imax)) {  //balayage de toutes les entrees de
        strcpy(chtemp, entrcour->id);     //la section
        strcat(chtemp, "=");
        strcat(chtemp, entrcour->val);
        AjouteCh(dest, chtemp, i, imax);
        entrcour = entrcour->suivt;
      }
      dest[i] = 0x00;  //double zero terminal
    }
    //  b) Recherche de l'entree dans la section
    else {
      trouve = 0;
      entrcour = sectcour->entr;
      while ((entrcour) && (!trouve))
        if (entrcour->comment)
          entrcour = entrcour->suivt;
        else
          if (!stricomp(entrcour->id, entr))
            trouve = 1;
          else
            entrcour = entrcour->suivt;
      // Entree non trouvee --> copie de la valeur par defaut
      if (!trouve)
        AjouteCh(dest, defaut, i, dim);
      // Entree trouvee --> recopie de la valeur
      else
        AjouteCh(dest, entrcour->val, i, dim);
    }
  return i-1;  //renvoi du nb de car. lus (sans le zero terminal)
}


//--- Lecture d'une entree sous forme d'un nombre entier dans le profil --------
int FichProf::LireInt(char *sect,   //nom de la section a lire
                      char *entr,   //nom de l'entree a lire
                      int  defaut)  //valeur par defaut
{
  int trouve=0;

  // Recherche de la section dans la liste chainee
  sectcour = profil;
  while ((sectcour) && (!trouve))
    if (sectcour->comment)
      sectcour = sectcour->suivt;
    else
      if (!stricomp(sectcour->id, sect))
        trouve = 1;
      else
        sectcour = sectcour->suivt;
  // Section non trouvee --> retour de la valeur par defaut
  if (!trouve)
    return defaut;
  // Recherche de l'entree dans la section
  trouve = 0;
  entrcour = sectcour->entr;
  while ((entrcour) && (!trouve))
    if (entrcour->comment)
      entrcour = entrcour->suivt;
    else
      if (!stricomp(entrcour->id, entr))
        trouve = 1;
      else
        entrcour = entrcour->suivt;
      // Entree non trouvee --> retour de la valeur par defaut
  if (!trouve)
    return defaut;
  // Entree trouvee --> Renvoie la valeur apres conversion
  else
    return atoi(entrcour->val);
}


//--- Ecriture d'une entree dans le profil -------------------------------------
int FichProf::EcrireDonnees(char *sect,  //nom de la section a mettre a jour
                            char *entr,  //nom de l'entree a mettre a jour
                            char *val)   //valeur a mettre a jour
{
  modif = 1;  //armement de l'indicateur de modification
  //---- Cas d'un profil vide : creation d'une section
  if (!profil)
    profil = NouvelleSection(sect, entr, val);
  //---- Mise a jour d'un profil existant
  else
    MajProfil(sect, entr, val);
  return 1;
}


//--- Lecture du profil sur disque et stockage en memoire ----------------------
int FichProf::LireProfil(void)
{
  if (!Ouvrir(0)) return 0;
  while (LireLgn());
  Fermer();
  modif = 0;  // desactivation de l'indicateur de modification
  return 1;
}


//--- Ecriture du profil sur disque --------------------------------------------
int FichProf::EcrireProfil(void)
{
  sectcour = profil;
  //Effacement du fichier precedent et creation du nouveau fichier
  Detruire();
  Creer();
  if (!Ouvrir(1)) return 0;
  //Ecriture des sections
  while (sectcour) {
    if (!EcrireSection()) return 0;
    sectcour = sectcour->suivt;
  }
  Fermer();
  modif = 0;  // desactivation de l'indicateur de modification
  return 1;
}


//--- Lecture d'une ligne dans le fichier de profil ----------------------------
int FichProf::LireLgn(void)
{
  char lgn[256];
  char *p;
  int res;

  // Lecture d'une ligne dans le fichier
  res = FichTxt::LireLgn(lgn);
  // Test de fin de fichier (eof rencontre et aucun caractere lu)
  if ((!res) && (!lgn[0]))
    return 0;
  // Traitement de la ligne
  p = lgn;
  while (*p == ' ') p++;  //On saute les espaces en debut de ligne
  if (!*p) return 1;      //Ligne vide
  if (*p == ';')
    return LireCommentaire(p);   //Ligne de commentaires
  else
    if (*p == '[')
      return LireSection(p);     //Ligne de titre de section
    else
      return LireEntree(p);      //Ligne d'entree de section
}


//--- Lecture d'une ligne d'entete de section ----------------------------------
int FichProf::LireSection(char *ch1)
{
  char ch2[256];
  char *p1 = ch1;
  char *p2 = ch2;

  // Extraction du nom de la section
  p1++;        //on saute le crochet ouvrant
  while ((*p1) && (*p1 != ']'))
    *p2++ = *p1++;
  if (p2 == ch2) return 0;  //sortie si aucun caractere a lire
  *p2 = 0x00;
  // Creation de la nouvelle section en memoire
  if (!profil) {
    profil = new Ssection;
    sectcour = profil;
  }
  else {
    sectcour->suivt = new Ssection;
    sectcour = sectcour->suivt;
  }
  // Stockage des informations
  sectcour->comment = 0;
  sectcour->id = new char[strlen(ch2)+1];
  strcpy(sectcour->id, ch2);
  sectcour->entr = NULL;
  sectcour->suivt = NULL;
  return 1;
}


//--- Lecture d'une ligne d'entree ---------------------------------------------
int FichProf::LireEntree(char *chlgn)
{
  char chid[256], chval[256];
  char *p1 = chlgn;  //pointeurs vers les chaines source et destination
  char *p2;

  // Lecture de l'identifiant
  p2 = chid;
  while ((*p1) && (*p1 != '='))
    *p2++ = *p1++;
  if (!*p1)
    return 0;  //sortie si ligne incomplete
  *p2 = 0x00;  //zero de fin de chaine
  *p1++;       //on saute le caractere '='
  // Lecture de la valeur
  p2 = chval;
  while (*p1)
    *p2++ = *p1++;
  *p2 = 0x00;  //zero de fin de chaine
  // Creation de la nouvelle entree
  if (!sectcour->entr) {   //cas d'une premiere entree dans la section
    sectcour->entr = new Sentree;
    entrcour = sectcour->entr;
  }
  else {                   //cas des entrees suivantes
    entrcour->suivt = new Sentree;
    entrcour = entrcour->suivt;
  }
  // Stockage des donnees lues
  entrcour->comment = 0;
  entrcour->id = new char[strlen(chid)+1];
  strcpy(entrcour->id, chid);
  entrcour->val = new char[strlen(chval)+1];
  strcpy(entrcour->val, chval);
  entrcour->suivt = NULL;
  return 1;
}


//--- Lecture d'une ligne de commentaires --------------------------------------
int FichProf::LireCommentaire(char *ch)
{
  int commentsection = 1;  //indique si le commentaire doit etre traite comme
                           //une section (commentaires en debut de fichier)

  // Creation d'une nouvelle section
  if (!profil) {           //cas de la premiere ligne du fichier
    profil = new Ssection;
    sectcour = profil;
  }
  else if (sectcour->comment) {  //cas des autres lignes de commentaire en debut
         sectcour->suivt = new Ssection;                            //de fichier
         sectcour = sectcour->suivt;
       }
       else                      //cas d'un commentaire de niveau entree
         commentsection = 0;
  // Stockage des donnees au niveau section
  if (commentsection) {
    sectcour->comment = 1;
    sectcour->id = new char[strlen(ch)+1];
    strcpy(sectcour->id, ch);
    sectcour->entr = NULL;
    sectcour->suivt = NULL;
  }
  // Creation d'une nouvelle entree
  else {
    if (!sectcour->entr) {   // - cas de la premiere entree
      sectcour->entr = new Sentree;
      entrcour = sectcour->entr;
    }
    else {                   // - cas des autres entrees
      entrcour->suivt = new Sentree;
      entrcour = entrcour->suivt;
    }
    // Stockage des donnees au niveau entree
    entrcour->comment = 1;
    entrcour->id = new char[strlen(ch)+1];
    strcpy(entrcour->id, ch);
    entrcour->suivt = NULL;
  }
  return 1;
}


//--- Ecrire d'une ligne d'entete de section dans le fichier de profil ---------
int FichProf::EcrireSection(void)
{
  char lgn[256];

  if (sectcour->comment) {  //cas d'un commentaire
    if (!EcrireLgn(sectcour->id)) return 0;
  }
  else {
    // Ecriture de l'entete de section
    strcpy(lgn, "[");
    strcat(lgn, sectcour->id);
    strcat(lgn, "]");
    if (!EcrireLgn(lgn)) return 0;
    entrcour = sectcour->entr;
    // Ecriture des entrees
    while (entrcour) {
      EcrireEntree();
      entrcour = entrcour->suivt;
    }
    EcrireLgn("");
  }
  return 1;
}


//--- Ecrire d'une ligne d'entree dans le fichier de profil --------------------
int FichProf::EcrireEntree(void)
{
  char lgn[256];

  if (entrcour->comment) {  //cas d'un commentaire
    if (!EcrireLgn(entrcour->id)) return 0;
  }
  else {
    strcpy(lgn, entrcour->id);
    strcat(lgn, "=");
    strcat(lgn, entrcour->val);
    if (!EcrireLgn(lgn)) return 0;
  }
  return 1;
}

//--- Mise a jour du profil avec de nouvelles donnees --------------------------
void FichProf::MajProfil(char *sect, char *entr, char *val)
{
  int trouve=0;
  Ssection *sectprec;

  // Recherche de la section dans la liste chainee
  sectprec = sectcour = profil;
  while ((sectcour) && (!trouve))
    if (sectcour->comment) {
      sectprec = sectcour;
      sectcour = sectcour->suivt;
    }
    else
      if (!stricomp(sectcour->id, sect))
        trouve = 1;
      else {
        sectprec = sectcour;
        sectcour = sectcour->suivt;
      }
  // Section non trouvee --> creation d'une nouvelle section
  if (!trouve)
    sectprec->suivt = NouvelleSection(sect, entr, val);
  // Section trouvee
  else
    // 1. Si entr est defini --> mise a jour de la section
    if (entr)
      MajSection(sectcour, entr, val);
    // 2. Si entr = NULL, on detruit la section.
    else {
      if (sectprec == profil)  //suppression du 1 element de la liste
        profil = sectcour->suivt;
      else                     //autres cas
        sectprec->suivt = sectcour->suivt;
      DesallouerSection(sectcour);
    }
}


//--- Mise a jour d'une section avec de nouvelles donnees ----------------------
void FichProf::MajSection(Ssection* sect, char *entr, char *val)
{
  int trouve=0;
  Sentree *entrprec;

  // Recherche de la section dans la liste chainee
  entrcour = entrprec = sect->entr;
  while ((entrcour) && (!trouve))
    if (entrcour->comment) {
      entrprec = entrcour;
      entrcour = entrcour->suivt;
    }
    else
      if (!stricomp(entrcour->id, entr))
        trouve = 1;
      else {
        entrprec = entrcour;
        entrcour = entrcour->suivt;
      }
  // Entree non trouvee --> creation d'une nouvelle entree
  if (!trouve)
    entrprec->suivt = NouvelleEntree(entr, val);
  // Entree trouvee --> mise a jour
  else
    // a) Si val est defini, on change la valeur
    if (val)
      MajEntree(entrcour, val);
    // b) Si val=NULL, on detruit l'entree
    else {
      if (entrprec == sect->entr)  //suppression du 1 element de la liste
        sect->entr = entrcour->suivt;
      else                         //autres cas
        entrprec->suivt = entrcour->suivt;
      DesallouerEntree(entrcour);
       }
}


//--- Mise a jour d'une entree -------------------------------------------------
void FichProf::MajEntree(Sentree* entr, char *val)
{
  delete entr->val;
  entr->val = new char[strlen(val)+1];
  strcpy(entr->val, val);
}


//--- Creation d'une nouvelle section ------------------------------------------
FichProf::Ssection *FichProf::NouvelleSection(char *sect, char *entr, char *val)
{
  Ssection *s;

  // Creation de la nouvelle section
  s = new Ssection;
  s->comment = 0;
  s->id = new char[strlen(sect)+1];
  strcpy(s->id, sect);
  s->suivt = NULL;
  // Creation de la nouvelle entree
  if (entr) {
    s->entr = new Sentree;
    s->entr->comment = 0;
    s->entr->id = new char[strlen(entr)+1];
    strcpy(s->entr->id, entr);
    s->entr->val = new char[strlen(val)+1];
    strcpy(s->entr->val, val);
    s->entr->suivt = NULL;
  }
  // Cas d'une section sans entree
  else
    s->entr = NULL;
  return s;
}


//--- Creation d'une nouvelle entree -------------------------------------------
FichProf::Sentree *FichProf::NouvelleEntree(char *entr, char *val)
{
  Sentree *e;

  e = new Sentree;
  e->comment = 0;
  e->id = new char[strlen(entr)+1];
  strcpy(e->id, entr);
  e->val = new char[strlen(val)+1];
  strcpy(e->val, val);
  e->suivt = NULL;
  return e;
}


//--- Ajoute une chaine de car. a la chaine destination ------------------------
void FichProf::AjouteCh(char *dest, char *source, int &i, int imax)
{
  int max = imax - 1;   //indice maxi dans dest (sans le zero terminal)
  char *p1 = dest + i;  //pointeurs vers les chaines source et destination
  char *p2 = source;

  while ((*p2) && (i < max)) {
    *p1++ = *p2++;
    i++;
  }
  *p1 = 0x00;   // zero terminal
  i++;
}


//--- Desallocation memoire du profil ------------------------------------------
void FichProf::DesallouerProfil(void)
{
  while (profil) {
    sectcour = profil->suivt;
    DesallouerSection(profil);
    profil = sectcour;
  }
}


//--- Desallocation memoire d'une section --------------------------------------
void FichProf::DesallouerSection(Ssection *sect)
{
  delete sect->id;
  while (sect->entr) {
    entrcour = sect->entr->suivt;
    DesallouerEntree(sect->entr);
    sect->entr = entrcour;
  }
  delete sect;
}


//--- Desallocation memoire d'une entree ---------------------------------------
void FichProf::DesallouerEntree(Sentree *entr)
{
  delete entr->id;
  if (!entr->comment)
    delete entr->val;
  delete entr;
}



//===== Fonctions d'acces aux fichiers de profil ===============================

int WritePrivateProfileString(char *section,  //section a mettre a jour
                              char *entree,   //entree a mettre a jour
                              char *source,   //chaine source
                              char *nomfich)  //nom du fichier de profil (avec chemin)
{
  FichProf FP(nomfich);

  return FP.EcrireDonnees(section, entree, source);
}


int GetPrivateProfileString(char *section,  //section a consulter
                            char *entree,   //entree a consulter
                            char *defaut,   //valeur par defaut
                            char *dest,     //chaine destination
                            int tdest,      //taille maxi de dest (avec zero terminal)
                            char *nomfich)  //nom du fichier de profil (avec chemin)
{
  FichProf FP(nomfich);

  return FP.LireCh(section, entree, defaut, dest, tdest);
}


int GetPrivateProfileInt(char *section,  //section a consulter
                         char *entree,   //entree a consulter
                         int defaut,     //valeur par defaut
                         char *nomfich)  //nom du fichier de profil (avec chemin)
{
  FichProf FP(nomfich);

  return FP.LireInt(section, entree, defaut);
}



//=== Routines diverses ========================================================

int stricomp(const char *ch1, const char *ch2)
{
  char c1, c2;

  while (*ch1 && *ch2) {
    // Conversion en majuscules des deux caracteres a comparer
    c1 = ((*ch1 >= 'a') && (*ch1 <= 'z')) ? *ch1 - 32 : *ch1;
    c2 = ((*ch2 >= 'a') && (*ch2 <= 'z')) ? *ch2 - 32 : *ch2;
    // Arret si caracteres differents
    if (c1 != c2) break;
    ch1++;
    ch2++;
  }
  // Determine la valeur de retour
  if (*ch1 < *ch2) return -1;
  if (*ch1 > *ch2) return  1;
  return 0;
}

